#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

## Copyright 2004/2005 by LivingLogic AG, Bayreuth/Germany.
## Copyright 2004/2005 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


"""
<par><module>ll.orasql</module> contains utilities for working with
<link href="http://www.computronix.com/utilities.shtml#Oracle"><app>cx_Oracle</app></link>:</par>
<ulist>
<item>It allows calling procedure with keyword arguments (via the
<pyref class="Proc"><class>Proc</class></pyref> class).</item>
<item>Query results will be put into <pyref class="Record"><class>Record</class></pyref>
objects, where database fields are accessible as object attributes.</item>
<item>The <pyref class="Cursor"><class>Cursor</class></pyref> class
provides methods for iterating through the database metadata.</item>
</ulist>
"""


import datetime

from cx_Oracle import *

from ll import misc


class SQLObjectNotfoundError(Exception):
	def __init__(self, obj):
		self.obj = obj

	def __str__(self):
		if self.obj.owner is not None:
			return "no %s named %r found for owner %r" % (self.obj.__class__.type, self.obj.name, self.obj.owner)
		else:
			return "no %s named %r found" % (self.obj.__class__.type, self.obj.name)


class UnknownModeError(ValueError):
	def __init__(self, mode):
		self.mode = mode

	def __str__(self):
		return "unknown mode %r" % self.mode


class Proc:
	"""
	A <class>Proc</class> object is a wrapper for calling a specific Oracle
	procedure with keyword arguments.
	"""
	_ora2cx = {
		"date": DATETIME,
		"number": NUMBER,
		"varchar2": STRING,
		"clob": CLOB,
		"blob": BLOB,
	}
	
	def __init__(self, name):
		"""
		Create a <class>Proc</class> object. <arg>name</arg> is the name of the
		Oracle procedure.
		"""
		self.name = name
		self._realargs = None

	def __repr__(self):
		return "Proc(%r)" % self.name

	def _calcrealargs(self, cursor):
		if self._realargs is None:
			self._realargs = ({}, [])
			cursor.execute("select lower(argument_name) as name, lower(in_out) as in_out, lower(data_type) as datatype from user_arguments where lower(object_name)=lower(:name) order by sequence", name=self.name)
			for (i, record) in enumerate(cursor):
				self._realargs[1].append((record.name, record.datatype, "in" in record.in_out, "out" in record.in_out))
				self._realargs[0][record.name] = (i, record.datatype, "in" in record.in_out, "out" in record.in_out)

	def __call__(self, cursor, **kwargs):
		"""
		Call the procedure with keyword arguments <arg>kwargs</arg>. <arg>cursor</arg>
		must be a <module>cx_Oracle</module> cursor. This will return a
		<pyref class="Record"><class>Record</class></pyref> object containing the result
		of the call.
		"""
		self._calcrealargs(cursor)
		realargs = [None]*len(self._realargs[1])
		for (key, value) in kwargs.iteritems():
			(pos, datatype, isin, isout) = self._realargs[0][key.lower()]
			if isinstance(value, unicode):
				value = value.encode("iso-8859-1")
			realargs[pos] = value

		# Replace out parameters (and strings that are longer than the allowed
		# maximum) with variables
		for (pos, (name, datatype, isin, isout)) in enumerate(self._realargs[1]):
			realarg = realargs[pos]
			if isout or (isinstance(realarg, str) and len(realarg) >= 32768):
				var = cursor.var(self._ora2cx[datatype])
				var.setvalue(0, realarg)
				realargs[pos] = var

		result = Record()
		for (i, value) in enumerate(cursor.callproc(self.name, realargs)):
			result[self._realargs[1][i][0]] = value
		return result


class LLProc(Proc):
	def __call__(self, cursor, **kwargs):
		args = {}
		for (key, value) in kwargs.iteritems():
			if key == "user":
				key = "c_user"
			else:
				key = "p_%s" % key
			args[key] = value

		result = Record()
		for (key, value) in Proc.__call__(self, cursor, **args).iteritems():
			if key == "c_user":
				key = "user"
			elif key.startswith("p_"):
				key = key[2:]
			else:
				raise ValueError("unknown parameter name %r in result" % key)
			result[key] = value
		return result


class Record(dict):
	"""
	A <class>Record</class> is a subclass of <class>dict</class> that is used
	for storing results of database queries. Both item and attribute access (i.e.
	<method>__getitem__</method> and <method>__getattr__</method>) are available.
	Field names are case insensitive.
	"""
	def fromdata(cls, cursor, row):
		"""
		This class method can be used to create a <class>Record</class> instance
		from the database data.
		"""
		record = cls()
		for (i, field) in enumerate(row):
			record[cursor.description[i][0]] = field
		return record
	fromdata = classmethod(fromdata)

	def __getitem__(self, name):
		return dict.__getitem__(self, name.lower())

	def __setitem__(self, name, value):
		dict.__setitem__(self, name.lower(), value)

	def __delitem__(self, name):
		dict.__delitem__(self, name.lower())

	def __getattr__(self, name):
		return self.__getitem__(name)

	def __setattr__(self, name, value):
		self.__setitem__(name, value)

	def __delattr__(self, name):
		self.__delitem__(name)

	def __repr__(self):
		return "%s.%s(%s)" % (self.__class__.__module__, self.__class__.__name__, ", ".join(["%s=%r" % item for item in self.iteritems()]))


class Connection(Connection):
	def cursor(self):
		return Cursor(self)


def connect(*args, **kwargs):
	return Connection(*args, **kwargs)


class Cursor(Cursor):
	"""
	A subclass of <module>cx_Oracle</module>s cursor class. All database results
	will be returned as <pyref class="Record"><class>Record</class> objects.
	"""
	def fetchone(self, type=Record):
		row = super(Cursor, self).fetchone()
		if row is not None:
			row = type.fromdata(self, row)
		return row

	def fetchmany(self, rows=0, type=Record):
		sup = super(Cursor, self)
		return [ type.fromdata(self, row) for row in sup.fetchmany(rows)]

	def fetchall(self, type=Record):
		sup = super(Cursor, self)
		return [ type.fromdata(self, row) for row in sup.fetchall()]

	def fetch(self, type=Record):
		while True:
			yield type.fromdata(self, self.next())

	def __iter__(self):
		return self.fetch()

	def iterobjects(self, mode="create", all=False):
		"""
		<par>Generator that yields the definition of all or the current users
		sequences, tables, primary keys, foreign keys, comments, unique constraints,
		views, functions, procedures, packages and type in topological order (i.e.
		if the definition of an object (e.g. a table) is generated it will only
		depend on objects whose definition has been generated before).</par>
		<par>If <arg>mode</arg> is <lit>"drop"</lit>, the order will be reversed.
		If <arg>all</arg> is false only the current users objects will be generated
		(except if the depend on objects from another user), else all objects
		will be generated.</par>
		"""
		if mode not in ("create", "drop"):
			raise UnknownModeError(mode)

		done = set()
	
		def dosequences():
			# select * from all_sequences where sequence_owner=nvl(:owner, user) and sequence_name=:name
			if all:
				self.execute("select decode(sequence_owner, user, null, sequence_owner) as sequence_owner, sequence_name from all_sequences")
			else:
				self.execute("select null as sequence_owner, sequence_name from user_sequences")
			for rec in self.fetchall():
				for definition in SequenceDefinition(rec.sequence_name, rec.sequence_owner).iterdependent(self, done, mode):
					yield definition

		def dotables():
			if all:
				self.execute("select decode(owner, user, null, owner) as owner, table_name from all_tables")
			else:
				self.execute("select null as owner, table_name from user_tables")
			for rec in self.fetchall():
				if mode=="create":
					for definition in TableDefinition(rec.table_name, rec.owner).iterdependent(self, done, mode):
						yield definition
		
				# Primary key
				if all:
					self.execute("select decode(owner, user, null, owner) as owner, constraint_name from all_constraints where constraint_type='P' and owner=:owner and table_name=:name", owner=rec.owner, name=rec.table_name)
				else:
					self.execute("select null as owner, constraint_name from user_constraints where constraint_type='P' and table_name=:name", name=rec.table_name)
				for rec2 in self.fetchall():
					for definition in PKDefinition(rec2.constraint_name).iterdependent(self, done, mode):
						yield definition

				# Comments
				for definition in CommentsDefinition(rec.table_name, rec.owner).iterdependent(self, done, mode):
					yield definition
	
				if mode=="drop":
					for definition in TableDefinition(rec.table_name, rec.owner).iterdependent(self, done, mode):
						yield definition

		def doconstraints():
			if all:
				self.execute("select constraint_type, decode(owner, user, null, owner) as owner, constraint_name from all_constraints where constraint_type in ('R', 'U')")
			else:
				self.execute("select constraint_type, null as owner, constraint_name from user_constraints where constraint_type in ('R', 'U')")
			types = {"U": UniqueDefinition, "R": FKDefinition}
			for rec in self.fetchall():
				for definition in types[rec.constraint_type](rec.constraint_name, rec.owner).iterdependent(self, done, mode):
					yield definition

		def dosynonyms():
			if all:
				self.execute("select decode(owner, user, null, owner) as owner, synonym_name from all_synonyms")
			else:
				self.execute("select null as owner, synonym_name from user_synonyms")
			for rec in self.fetchall():
				for definition in SynonymDefinition(rec.synonym_name, rec.owner).iterdependent(self, done, mode):
					yield definition

		def doviews():
			if all:
				self.execute("select decode(owner, user, null, owner) as owner, view_name from all_views")
			else:
				self.execute("select null as owner, view_name from user_views")
			for rec in self.fetchall():
				for definition in ViewDefinition(rec.view_name, rec.owner).iterdependent(self, done, mode):
					yield definition

		def docode():
			for type in (FunctionDefinition, ProcedureDefinition, PackageDefinition, TypeDefinition, TriggerDefinition):
				if all:
					self.execute("select decode(owner, user, null, owner) as owner, object_name from all_objects where lower(object_type)=lower(:type)", type=type.type)
				else:
					self.execute("select null as owner, object_name from user_objects where lower(object_type)=lower(:type)", type=type.type)
				for rec in self.fetchall():
					for definition in type(rec.object_name).iterdependent(self, done, mode):
						yield definition

		funcs = (dosequences, dotables, doconstraints, dosynonyms, doviews, docode)
		if mode=="drop":
			funcs = reversed(funcs)
		for func in funcs:
			for definition in func():
				yield definition


def formatstring(value):
	result = []
	current = []

	# Helper function: move the content of current to result
	def shipcurrent(force=False):
		if current and (force or (len(current) > 2000)):
			if result:
				result.append(" || ")
			result.append("'%s'" % "".join(current))

	for c in value:
		if c == "'":
			current.append("''")
			shipcurrent()
		elif ord(c) < 32 or ord(c)>127:
			shipcurrent(True)
			current = []
			if result:
				result.append(" || ")
			result.append("chr(%d)" % ord(c))
		else:
			current.append(c)
			shipcurrent()
	shipcurrent(True)
	return "".join(result)


class Definition(object):
	name2type = {}

	class __metaclass__(type):
		def __new__(mcl, name, bases, dict):
			typename = None
			if "type" in dict and name != "Definition":
				typename = dict["type"]
			cls = type.__new__(mcl, name, bases, dict)
			if typename is not None:
				Definition.name2type[typename] = cls
			return cls

	def __init__(self, name, owner=None):
		self.name = name
		self.owner = owner

	def __repr__(self):
		if self.owner is not None:
			return "%s.%s(%r, %r)" % (self.__class__.__module__, self.__class__.__name__, self.name, self.owner)
		else:
			return "%s.%s(%r)" % (self.__class__.__module__, self.__class__.__name__, self.name)

	def __eq__(self, other):
		return self.__class__ is other.__class__ and self.name == other.name and self.owner == other.owner

	def __ne__(self, other):
		return not self.__eq__(other)

	def __hash__(self):
		return hash(self.__class__.__name__) ^ hash(self.name) ^ hash(self.owner)

	@misc.notimplemented
	def createddl(self, cursor):
		"""
		Return an &sql; statement that will recreate <self/> in the database.
		"""

	@misc.notimplemented
	def dropddl(self, cursor):
		"""
		Return an &sql; statement that will delete <self/> in the database.
		"""

	def iterdependent(self, cursor, done=None, mode="create"):
		if done is None:
			done = set()
		if self not in done:
			if mode=="drop":
				cursor.execute("select type, decode(owner, user, null, owner) as owner, name from all_dependencies where lower(referenced_type)=:type and referenced_name=:name and referenced_owner=nvl(:owner, user)", type=self.__class__.type, name=self.name, owner=self.owner)
			elif mode=="create":
				cursor.execute("select referenced_type as type, decode(referenced_owner, user, null, referenced_owner) as owner, referenced_name as name from all_dependencies where lower(type)=:type and name=:name and owner=nvl(:owner, user)", type=self.__class__.type, name=self.name, owner=self.owner)
			else:
				raise UnknownModeError(mode)
			for rec in cursor.fetchall():
				if rec.type != "NON-EXISTENT":
					for definition in Definition.name2type[rec.type.lower()](rec.name, rec.owner).iterdependent(cursor, done, mode):
						yield definition
			done.add(self)
			yield self


class SequenceDefinition(Definition):
	type = "sequence"

	def createddl(self, cursor):
		cursor.execute("select * from all_sequences where sequence_owner=nvl(:owner, user) and sequence_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			code  = "create sequence %s.%s\n" % (rec.sequence_owner, rec.sequence_name)
		else:
			code  = "create sequence %s\n" % rec.sequence_name
		code += "\tincrement by %d\n" % rec.increment_by
		code += "\tstart with %d\n" % rec.min_value
		code += "\tmaxvalue %s\n" % rec.max_value
		code += "\tminvalue %d\n" % rec.min_value
		code += "\t%scycle\n" % ["no", ""][rec.cycle_flag=='Y']
		code += "\tcache %d\n" % rec.cache_size
		code += "\t%sorder;\n" % ["no", ""][rec.order_flag=='Y']
		return code

	def dropddl(self, cursor):
		if self.owner is not None:
			return "drop sequence %s.%s" % (self.owner, self.name)
		else:
			return "drop sequence %s" % self.name


class TableDefinition(Definition):
	type = "table"

	def createddl(self, cursor):
		cursor.execute("select * from all_tab_columns where owner=nvl(:owner, user) and table_name=:name order by column_id asc", owner=self.owner, name=self.name)
		recs = cursor.fetchall()
		if not recs:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			code = ["create table %s.%s\n(\n" % (self.owner, self.name)]
		else:
			code = ["create table %s\n(\n" % self.name]
		for rec in recs:
			fsize = rec.data_precision
			flen = rec.data_length
			fprec = rec.data_scale
			ftype = rec.data_type.lower()
			code.append("\t%s %s" % (rec.column_name.lower(), ftype))
			if rec.char_length:
				fsize = rec.char_length
			if fsize is not None:
				code.append("(%d" % fsize)
				if fprec is not None:
					code.append(", %d" % fprec)
				code.append(")")
			if rec.data_default is not None:
				code.append(" default %s" % rec.data_default)
			if rec.nullable == 'N':
				code.append(" not null")
			code.append(",\n")
		del code[-1]
		code.append("\n);\n")
		return "".join(code)

	def dropddl(self, cursor):
		if self.owner is not None:
			return "drop table %s.%s" % (self.owner, self.name)
		else:
			return "drop table %s" % self.name


class PKDefinition(Definition):
	type = "pk"

	def createddl(self, cursor):
		cursor.execute("select decode(owner, user, null, owner) as owner, constraint_name, table_name, r_owner, r_constraint_name from all_constraints where constraint_type='P' and owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec2 = cursor.fetchone()
		if rec2 is None:
			raise SQLObjectNotfoundError(self)
		cursor.execute("select column_name from all_cons_columns where owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		if rec2.owner is not None:
			name = "%s.%s" % (rec2.owner, rec2.table_name)
		else:
			name = rec2.table_name
		return "alter table %s add constraint %s primary key(%s);\n" % (name, self.name, ", ".join(r.column_name for r in cursor))

	def dropddl(self, cursor):
		cursor.execute("select decode(owner, user, null, owner) as owner, table_name from all_constraints where owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec.owner is not None:
			name = "%s.%s" % (rec.owner, rec.table_name)
		else:
			name = rec.table_name
		return "alter table %s drop constraint %s" % (name, self.name)


class CommentsDefinition(Definition):
	type = "comments"

	def createddl(self, cursor):
		cursor.execute("select column_name, comments from all_col_comments where owner=nvl(:owner, user) and table_name=:name and comments is not null", owner=self.owner, name=self.name)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, self.name)
		else:
			name = self.name
		return "\n".join("comment on column %s.%s is %s;" % (name, r.column_name, formatstring(r.comments)) for r in cursor) + "\n"

	def dropddl(self, cursor):
		# will be dropped with the table
		return ""


class FKDefinition(Definition):
	type = "fk"

	def createddl(self, cursor):
		# Add constraint_type to the query, so we don't pick up another constraint by accident
		cursor.execute("select decode(r_owner, user, null, r_owner) as r_owner, r_constraint_name, table_name from all_constraints where constraint_type='R' and owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		cursor.execute("select column_name from all_cons_columns where owner=nvl(:owner, user) and constraint_name=:name order by position", owner=self.owner, name=self.name)
		fields1 = ", ".join(r.column_name for r in cursor)
		cursor.execute("select table_name, column_name from all_cons_columns where owner=nvl(:owner, user) and constraint_name=:name order by position", owner=rec.r_owner, name=rec.r_constraint_name)
		if rec.r_owner is not None:
			fields2 = ", ".join("%s.%s(%s)" % (rec.r_owner, r.table_name, r.column_name) for r in cursor)
		else:
			fields2 = ", ".join("%s(%s)" % (r.table_name, r.column_name) for r in cursor)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, rec.table_name)
		else:
			name = rec.table_name
		return "alter table %s add constraint %s foreign key (%s) references %s;\n" % (name, self.name, fields1, fields2)

	def _ddl(self, cursor, cmd):
		cursor.execute("select table_name from all_constraints where owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, rec.table_name)
		else:
			name = rec.table_name
		return "alter table %s %s constraint %s" % (name, cmd, self.name)

	def dropddl(self, cursor):
		return self._ddl(cursor, "drop")

	def enableddl(self, cursor):
		return self._ddl(cursor, "disable")

	def disableddl(self, cursor):
		return self._ddl(cursor, "disable")


class UniqueDefinition(Definition):
	type = "unique"

	def createddl(self, cursor):
		# Add constraint_type to the query, so we don't pick up another constraint by accident
		cursor.execute("select table_name from all_constraints where constraint_type='U' and owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, rec.table_name)
		else:
			name = rec.table_name
		cursor.execute("select column_name from all_cons_columns where owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		return "alter table %s add constraint %s unique(%s);\n" % (name, self.name, ", ".join(r.column_name for r in cursor))

	def dropddl(self, cursor):
		cursor.execute("select table_name from all_constraints where constraint_type='U' and owner=nvl(:owner, user) and constraint_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, rec.table_name)
		else:
			name = rec.table_name
		return "alter table %s drop constraint %s" % (name, self.name)


class SynonymDefinition(Definition):
	type = "synonym"

	def createddl(self, cursor):
		cursor.execute("select table_owner, table_name from all_synonyms where owner=nvl(:owner, user) and synonym_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		owner = self.owner
		if owner == "PUBLIC":
			public = "public "
			owner = None
		else:
			public = ""
		if owner is not None:
			name = "%s.%s" % (owner, self.name)
		else:
			name = self.name

		if rec.table_owner is not None:
			name2 = "%s.%s" % (rec.table_owner, rec.table_name)
		else:
			name2 = rec.table_name
		return "create or replace %ssynonym %s for %s;\n" % (public, name, name2)

	def dropddl(self, cursor):
		owner = self.owner
		if owner == "PUBLIC":
			public = "public "
			owner = None
		else:
			public = ""
		if owner is not None:
			name = "%s.%s" % (owner, self.name)
		else:
			name = self.name
		return "drop %ssynonym %s" % (public, name)


class ViewDefinition(Definition):
	type = "view"

	def createddl(self, cursor):
		cursor.execute("select text from all_views where owner=nvl(:owner, user) and view_name=:name", owner=self.owner, name=self.name)
		rec = cursor.fetchone()
		if rec is None:
			raise SQLObjectNotfoundError(self)
		if self.owner is not None:
			name = "%s.%s" % (self.owner, self.name)
		else:
			name = self.name
		return "create or replace view %s as\n\t%s;\n/\n" % (self.name, rec.text.strip())

	def dropddl(self, cursor):
		if self.owner is not None:
			return "drop view %s.%s" % (self.owner, self.name)
		else:
			return "drop view %s" % self.name


class CodeDefinition(Definition):
	def createddl(self, cursor):
		cursor.execute("select text from all_source where lower(type)=lower(:type) and owner=nvl(:owner, user) and name=:name order by line", type=self.__class__.type, owner=self.owner, name=self.name)
		code = "".join(rec.text for rec in cursor)
		code = code.strip()
		if self.owner is not None:
			type = self.__class__.type
			code = code[code.lower().find(type)+len(type):].strip() # drop "procedure" etc.
			code = "create or replace %s %s.%s\n/\n" % (type, self.owner, code)
		else:
			code = "create or replace %s\n/\n" % code
		return code

	def dropddl(self, cursor):
		if self.owner is not None:
			return "drop %s %s.%s" % (self.__class__.type, self.owner, self.name)
		else:
			return "drop %s %s" % (self.__class__.type, self.name)


class FunctionDefinition(CodeDefinition):
	type = "function"


class ProcedureDefinition(CodeDefinition):
	type = "procedure"


class PackageDefinition(CodeDefinition):
	type = "package"


class TypeDefinition(CodeDefinition):
	type = "type"


class TriggerDefinition(CodeDefinition):
	type = "trigger"
